/*
 * RandomizedAdaptionProvider.java
 *
 * Created on 3. ervenec 2005, 21:31
 */

package cz.matfyz.sykora.nngui.data;

import cz.matfyz.sykora.nngui.exception.*;
import java.util.*;

/**
 * Provider slouc jako proxy mezi skutenm zdroem dat a jejich uivatelem.
 * Nate vechny zznamy ze zadanho provideru vstupnch dat a poadovanch
 * vstup a nsledn je v nhodnm poad pedv na podn uivateli.
 * Poet opakovn zznam nen nijak omezen, take tento provider lze pout
 * pro efektivn pedkldn ucch vzor sti.
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class RandomizedAdaptionProvider implements DataProvider {
    /**
     * Prv zpracovvan zznam z <i>samples</i> vybran pomoc genertoru
     * nhodnch sel bhem voln <i>fetchRow</i>.
     *  @see #fetchRow
     *  @see #inputPartSize
     *  @see #samples
     */
    protected double[] currentRow;
    /**
     * Seznam ucch vzork. Vstupn vzor i poadovan vstup jsou uloeny v
     * jednom poli, vstupn vzor je na indexech 0 a <i>inputPartSize</i> - 1,
     * nsledujc indexy v poli obsahuj oekvan vstup.
     *  @see #currentRow
     *  @see #initialize
     */
    protected Vector<double[]> samples;
    /**
     * Velikost vstupnho vzoru v polch ukldanch v <i>samples</i>. Pole
     * hodnot pak na prvnch <i>inputPartSize</i> indexech obsahuj vstupn vzor
     * a na zbvajcch jeho poadovan vstup.
     *  @see #samples
     */
    protected int inputPartSize;
    /**
     * Genertor nhodnch sel pouit pro vybrn pedkldanch vzork.
     *  @see #fetchRow
     */
    protected static Random randomGenerator = new Random(9876543);
    
    /**
     * Pomocn tda simulujc provider poadovanho vstupu k vzorm,
     * pedvanm pomoc tohoto providera. Vyuv struktury tohoto providera,
     * vetn pole <i>currentRow</i> ze kterho bere hodnoty pro aktuln
     * zpracovvn vzor z druh sti pole.
     */
    protected class DesiredOutputProvider implements DataProvider {
        /**
         * Uzaven providera. Nedl nic, uzaven nech na nadzen td.
         */
        public void close() {    
        }
        
        /**
         * Naten zznamu do providera. Pouze zkontroluje, jestli je v
         * nadzen naten poiteln zznam.
         *  @return true, pokud <i>currentRow</i> obsahuje pouiteln zznam,
         *              jinak vrac false
         *  @see RandomizedAdaptionProvider#currentRow
         */
        public boolean fetchRow() throws DataException {
            if(currentRow == null)
                return false;
            
            return true;
        }
        
        /**
         * Zjitn velikosti zznamu.
         *  @return velikost zznamu poadovanho vstupu
         *  @see RandomizedAdaptionProvider#inputPartSize
         *  @see RandomizedAdaptionProvider#samples
         */
        public int getRowSize() {
            if(samples.size() == 0)
                return 0;
            
            return samples.get(0).length - inputPartSize;
        }
        
        /**
         * Zjitn hodoty oekvanho vstupu na zadanm indexu.
         *  @param _index index hodnoty, na kterou se dotazujeme
         *  @return hodnota na index <i>_index</i>
         *  @throws DataProviderNotReadyException pokud jet nebyl do
         *              (nadazenho) providera naten dn zznam pomoc
         *              <i>fetchRow</i>
         *  @throws ArrayIndexOutOfBoundsException pokud <i>_index</i> nen
         *              platn index hodnoty v provideru
         *  @see RandomizedAdaptionProvider#currentRow
         *  @see RandomizedAdaptionProvider#inputPartSize
         */
        public double getValue(int _index) throws DataException {
            if(currentRow == null)
                throw new DataProviderNotReadyException(this);
            if(_index < 0 || _index >= currentRow.length - inputPartSize)
                throw new ArrayIndexOutOfBoundsException(_index);
            
            return currentRow[inputPartSize + _index];
        }
        
        /**
         * Petoen zznam v provideru na zatek. Vdy uspje, stejne jako
         * <i>rewind</i> nadazen tdy.
         *  @return true za vech okolnost
         */
        public boolean rewind() {
            return true;
        }
    }
    
    /**
     * Uzaven tohoto provider. Pitom pro sporu pamti vyprzdn seznam
     * natench vzor.
     *  @see #samples
     */
    public void close() {
        samples.clear();
    }
    
    /**
     * Naten dalho zznamu z providera. Piprav pro ten nhodn vybran
     * zznam z natench bhem inicializace. Pokud je seznam vzor neprzdn,
     * vdy uspje.
     *  @return true pokud je seznam natench zznam neprzdn, jinak false.
     *  @see #currentRow
     *  @see #samples
     */
    public boolean fetchRow() throws DataException {
        if(samples.size() == 0)
            return false;
        
        int new_index = randomGenerator.nextInt(samples.size());
        currentRow = samples.get(new_index);
        
        return true;
    }
    
    /**
     * Funkce vytvejc provider po poadovan vstupy k aktuln natenm
     * zznamm.
     *  @return novou instanci providera pro pstup k poadovanm vstupm
     *              k aktuln natenm vzorm.
     *  @see DesiredOutputProvider
     */
    public DataProvider getDesiredOutputProvider() {
        return new DesiredOutputProvider();
    }
    
    /**
     * Zjitn velikosti zznamu poskytovanho tmto providerem.
     *  @return velikost zznamu poskytovanho tmto providerem
     *  @see #inputPartSize
     */
    public int getRowSize() {
        return inputPartSize;
    }
    
    /**
     * Zjitn seln hodnoty na zadanm indexu.
     *  @param _index index hodnoty na kterou se dotazujeme
     *  @return hodnota na indexu <i>_index</i>
     *  @throws DataProviderNotReadyException pokud dosud nebyl naten dn
     *              zznam pomoc <i>fetchRow</i>
     *  @throws ArrayIndexOutOfBoundsException pokud <i>_index</i> nen platnm
     *              indexem hodnoty
     *  @see #currentRow
     */
    public double getValue(int _index) throws DataException {
        if(currentRow == null)
            throw new DataProviderNotReadyException(this);
        if(_index >= inputPartSize || _index < 0)
            throw new ArrayIndexOutOfBoundsException(_index);
        
        return currentRow[_index];
    }
    
    /**
     * Inicializace providera. Nat z provider data, dokud njak obsahuj
     * (pozor na providery jako je tento, ze kterch jde natat data
     * donekonena!).
     *  @param _input provider, ze kterho jsou natna vstupn vzory
     *  @param _desired_output provider, ze kterho jsou natny poadovan
     *              vstupy k vstupnm vzorm. Poadovan vstup je natn hned
     *              po naten vstupnho vzoru, pedpokld se, e poadovan
     *              vstup odpovd vstupnmu vzoru, kter byl naten ped nm.
     *  @see #inputPartSize
     *  @see #samples
     */
    protected void initialize(DataProvider _input,DataProvider _desired_output) {
        assert(_input != null);
        assert(_desired_output != null);
        
        try {
            // pokud jsou providery prazdne, rovnou skonci a nic nealokuj
            if(!_input.fetchRow() || !_desired_output.fetchRow())
                return;
            
            samples = new Vector<double[]>();
            inputPartSize = _input.getRowSize();
            
            int sample_size = _input.getRowSize() + _desired_output.getRowSize();
            
            do {
                double[] row = new double[sample_size];
                for(int i = 0; i < inputPartSize; i++)
                    row[i] = _input.getValue(i);
                for(int i = 0; i < sample_size - inputPartSize; i++)
                    row[inputPartSize + i] = _desired_output.getValue(i);
                samples.add(row);
            } while(_input.fetchRow() && _desired_output.fetchRow());
        }
        catch (DataException de) {
            // pokud doslo k vyjimce pri nacitani dat, vypis chybovou hlasku a
            // nacitani ukonci. Nactena data ale nezahazuj.
            System.err.println(de.getMessage());
            return;
        }
    }
    
    /**
     * Petoen dat v provideru zpt na zatek. Pokud seznam <i>samples</i>
     * obsahuje alespo jeden zznam, vdy projde. Nedl nic, dal zznamy
     * jsou opt pedkldny v nhodnm poad.
     *  @return true pokud v provideru je aten alespo jeden zznam, false
     *              pokud v provideru dn zznam naten nen.
     */
    public boolean rewind() {
        return samples.size() > 0;
    }
    
    /**
     * Konstruktor. Nate zznamy ze zadanch provider.
     *  @param _input provider, ze kterho se natou vstupn vzory
     *  @param _desired_output provider, ze kterho se natou poadovan vstupy
     *              k natenm vzorm.
     *  @see #initialize
     */
    public RandomizedAdaptionProvider(DataProvider _input,DataProvider _desired_output) {
        initialize(_input, _desired_output);
    }
}
