/*
 * TextStreamAdoptionProvider.java
 *
 * Created on 2. ervenec 2005, 20:34
 */

package cz.matfyz.sykora.nngui.data;

import cz.matfyz.sykora.nngui.exception.*;
import java.io.*;
import java.util.*;

/**
 * Implementace zdroje dat, kter poskytuje pstup nejen k testovacm datm,
 * ale souasn zpstupuje i poadovan vstup st. Ve se nat z jednoho
 * textovho souboru, data oekv v formtu 
 * <pre>[seln vstupy] => [oekvan vstupy]</pre>
 * vdy jeden zznam na dek
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class TextStreamAdoptionProvider implements DataProvider {
    
    protected double[] currentRow;
    
    protected double[] currentRowDesiredOutput;
    
    protected boolean hasCurrentRow;
    
    protected Reader inputReader;
    
    protected boolean prefetchedRow;
    
    protected StreamTokenizer tokenizer;
    
    protected class DesiredOutputProvider implements DataProvider {
        public void close() {
        }
        
        public boolean fetchRow() throws DataException {
            if(currentRowDesiredOutput == null)
                throw new DataProviderNotReadyException(this);
            
            return true;
        }
        
        public int getRowSize() {
            if(currentRowDesiredOutput == null)
                return 0;
            
            return currentRowDesiredOutput.length;
        }
        
        public double getValue(int _index) throws DataException {
            if(currentRowDesiredOutput == null)
                throw new DataProviderNotReadyException(this);
            
            return currentRowDesiredOutput[_index];
        }
        
        public boolean rewind() {
            return doRewind();
        }
    }
    
    public void close() {
        if(inputReader != null) {
            try {
                inputReader.close();
            }
            catch (IOException io) {
                 System.err.println("TextStreamProvider.close(): failed to close " +
                                   "inputStreamReader");
            }
            inputReader = null;
        }
    }
    
    private final boolean doRewind() {
        return rewind();
    }
    
    public boolean fetchRow() throws DataException {
        if(prefetchedRow) {
            prefetchedRow = false;
            return true;
        }
        
        try {
            if(currentRow == null) {
                Vector<Double> values = new Vector<Double>();

                tokenizer.nextToken();
                if(tokenizer.ttype == StreamTokenizer.TT_EOF)
                    return false;

                while(tokenizer.ttype == StreamTokenizer.TT_NUMBER) {
                    values.add(new Double(tokenizer.nval));
                    tokenizer.nextToken();
                }

                if(tokenizer.ttype != StreamTokenizer.TT_WORD)
                    throw new InvalidDataFormatException("There is no desired output " +
                                    "part in the data file");
                tokenizer.nextToken();

                if(tokenizer.ttype != StreamTokenizer.TT_NUMBER)
                    throw new InvalidDataFormatException("There must be at least one " +
                                    "number in the desired output part of the data file");

                Vector<Double> desired = new Vector<Double>();
                while(tokenizer.ttype == StreamTokenizer.TT_NUMBER) {
                    desired.add(new Double(tokenizer.nval));
                    tokenizer.nextToken();
                }

                if(tokenizer.ttype != StreamTokenizer.TT_EOL
                        && tokenizer.ttype != StreamTokenizer.TT_EOF)
                    throw new InvalidDataFormatException("Unexpected token. Number or" +
                                    " end of line/file expected");

                currentRow = new double[values.size()];
                currentRowDesiredOutput = new double[desired.size()];

                for(int i = 0; i < currentRow.length; i++)
                    currentRow[i] = values.get(i).doubleValue();
                for(int i = 0; i < currentRowDesiredOutput.length; i++)
                    currentRowDesiredOutput[i] = desired.get(i).doubleValue();
            }
            else {
                for (int i = 0; i < currentRow.length; i++) {
                    tokenizer.nextToken();
                    if(tokenizer.ttype == StreamTokenizer.TT_EOF)
                        return false;
                    else if(tokenizer.ttype != StreamTokenizer.TT_NUMBER)
                        throw new InvalidDataFormatException("There is not enough numbers " +
                                    "in the data part of a record");
                }

                if(tokenizer.ttype != StreamTokenizer.TT_WORD)
                    throw new InvalidDataFormatException("Data and desired output part " +
                                    "separator expected");

                for (int i = 0; i < currentRowDesiredOutput.length; i++) {
                    tokenizer.nextToken();
                    if(tokenizer.ttype != StreamTokenizer.TT_NUMBER)
                        throw new InvalidDataFormatException("There is not enough numbers " +
                                    "in the desired output section");
                }

                if(tokenizer.ttype != StreamTokenizer.TT_EOL 
                        && tokenizer.ttype != StreamTokenizer.TT_EOF)
                    throw new InvalidDataFormatException("The desired output section is " +
                                    "too wide");
            }
        }
        catch (IOException io) {
            return false;
        }
        
        return true;
    }
    
    public int getRowSize() {
        if(currentRow == null)
            return 0;
        
        return currentRow.length;
    }
    
    public double getValue(int _index) throws DataException {
        if(currentRow == null)
            throw new DataProviderNotReadyException(this);
        
        return currentRow[_index];
    }
    
    protected void initialize(Reader _input_reader) {
        assert(_input_reader != null);
        
        inputReader = _input_reader;
        tokenizer = new StreamTokenizer(inputReader);
        tokenizer.eolIsSignificant(true);
        tokenizer.parseNumbers();
        tokenizer.wordChars('=','=');
        tokenizer.wordChars('>','>');
    }
    
    public boolean rewind() {
        if(inputReader != null) {
            try {
                inputReader.reset();
            }
            catch (IOException io) {
                return false;
            }
            return true;
        }
        return false;
    }
    
    public TextStreamAdoptionProvider(Reader _input) {
    }
    
}
