/*
 * TextStreamConsumer.java
 *
 * Created on 2. ervenec 2005, 22:44
 */

package cz.matfyz.sykora.nngui.data;

import cz.matfyz.sykora.nngui.exception.*;
import java.io.*;

/**
 * Jednoduch konzument dat, kter to, co dostane zapisuje do textovho souboru.
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class TextStreamConsumer implements DataConsumer {
    /**
     * Pole, do kterho jsou ukldny doasn hodnoty mezi jejich pednm
     * pomoc <i>setValue</i> a potvrzenm obsahu dky pomoc <i>commitRow</i>.
     * Je alokovno pi nastaven velikosti zznamu ve vstupnm souboru, pokud
     * dosud alokovno nebylo, nebo jeho velikost neodpovd velikosti
     * poadovan uivatelem tohoto konzumenta.
     *  @see #setValue
     *  @see #commitRow
     *  @see #setRowSize
     */
    protected double[] currentRow;
    /**
     * Flag urujc, jestli se m pi zaven konzumenta zavt tak writer, do
     * kterho zapisuje data
     *  @see #close
     *  @see #output
     */
    protected boolean closeOutputOnClose;
    /**
     * Writer, do kterho jsou zapisovny poadovan dky.
     * Je inicializovn pi vytven konzmenta, pi uzaven konzumenta pomoc
     * <i>close</i> me bt uzaven, podle nastaven <i>closeOutputOnClose</i>
     *  @see #closeOutputOnClose
     *  @see #commitRow
     *  @see #initialize
     */
    protected PrintWriter output;
    
    /**
     * Uzave konzumenta.
     * Ped uzavenm vyprzdn buffer writeru pouitho pro zpis pedvanch
     * hodnot. Pokud je to nutn, zave vstupn writer
     *  @see #closeOutputOnClose
     */
    public void close() {
        output.flush();
        if(closeOutputOnClose) {
            output.close();
            output = null;
        }
    }
    
    /**
     * Potvrzen zznamu. Zape pedan zznam do vstupnho streamu jako dek
     * s elnm hodnotami oddlenmi mezerou.
     *  @return true pokud se zpis poda (tj. pokud existuje vstupn proud),
     *          jinak false.
     *  @see #currentRow
     */
    public boolean commitRow() {
        if(output == null)
            return false;
        
        for (int i = 0; i < currentRow.length; i++) {
            output.print(currentRow[i]);
            if(i < currentRow.length - 1)
                output.print(" ");
        }
        output.println();
        
        return true;
    }
    
    /**
     * Zjitn velikosti zznamu. Vrac velikost aktuln alokovanho bufferu
     * pro vstupn hodnoty.
     *  @return velikost bufferu pro vstupn hodnoty
     *  @see #currentRow
     */
    public int getRowSize() {
        if(currentRow == null)
            return 0;
        
        return currentRow.length;
    }
    
    /**
     * Inicializace providera. Poskytuje funkcionalitu spolenou pro vechny
     * konstruktory.
     *  @param _output writer, do kterho jsou zapisovny potvrzen dky.
     *  @param _close_output promnn urujc, jestli se m pi zaven
     *              providera pomoc <i>close</i> zavt i vstupn proud
     *              (true znamen zavt).
     *  @see #output
     *  @see #closeOutputOnClose
     */
    private void initialize(PrintWriter _output, boolean _close_output) {
        assert(_output != null);
        
        output = _output;
        closeOutputOnClose = _close_output;
    }
    
    /**
     * Nastaven velikosti zznamu pro providera. Pokud je nov velikost vt
     * ne aktuln velikost bufferu, alokuje nov, dostaten velk. Pi
     * nastaven velikosti zznamu providera mou bt zahozeny u pedan
     * hodnoty.
     *  @param _size nov nastavovan velikost zznamu konzumenta
     *  @return true, nastaven nov velikosti je vdy mon
     *  @throws IllegalArgumentException pokud nov nastavovan velikost nen
     *              platn velikost zznamu (tj. pokud neobsahuje alespo jednu
     *              hodnotu).
     */
    public boolean setRowSize(int _size) {
        if(_size <= 0)
            throw new IllegalArgumentException("Output row size must be greater than zero");
        
        // alokuj nove pole, pokud dosud zadne alokovano nebylo, nebo pokud jeho
        // soucasna velikost nedostacuje
        if(currentRow == null || currentRow.length < _size) {
            currentRow = new double[_size];
        }
        
        return true;
    }
    
    /**
     * Ped hodnotu na zadanm indexu.
     *  @param _index index pedvan hodnoty
     *  @param _value pedvan hodnota
     *  @throws DataConsumerNotReadyException pokud dosud v konzumentu nebyla
     *              nastavena velikost zznamu pomoc <i>setRowSize</i>
     *  @see #currentRow
     *  @see #setRowSize
     */
    public void setValue(int _index, double _value) throws DataException {
        if(currentRow == null)  
            throw new DataConsumerNotReadyException("Data consumer was not allocated");
        
        currentRow[_index] = _value;
    }
    
    /**
     * Konstruktor. Inicializuje konzumenta pro writer <i>_output</i>. Writer
     * nen pi zaven konzumenta uzaven.
     *  @param _output writer, do kterho jsou zapisovny potvrzen zznamy.
     *  @throws NullPointerException pokud je <i>_output</i> null.
     *  @see #initialize
     *  @see #TextStreamConsumer(java.io.PrintWriter, boolean)
     */
    public TextStreamConsumer(PrintWriter _output) {
        if(_output == null)
            throw new NullPointerException("Output writer must not be null");
        
        initialize(_output, false);
    }
    
    /**
     * Konstruktor. Inicializuje konzumenta pro writer <i>_output</i>,
     * parametrem <i>_close_output_on_close</i> lze manuln urit, jestli se m
     * writer zavt pi zaven konzumenta.
     *  @param _output writer, do kterho jsou zapisovny potvrzen zznamy.
     *  @param _close_output_on_close uruje, jestli se m writer zavt pi
     *              zaven konzumenta.
     *  @throws NullPointerException pokud je <i>_output</i> null.
     *  @see #initialize
     */
    public TextStreamConsumer(PrintWriter _output,boolean _close_output_on_close) {
        if(_output == null)
            throw new NullPointerException("Output writer must not be null");
        
        initialize(_output, _close_output_on_close);
    }
    
    /**
     * Konstruktor. Inicializuje konzumenta pro zzpis do souboru jmnem
     * <i>_output_file_name</i>. Vstup do souboru ja zaven pi zaven
     * konzumenta.
     *  @param _output_file_name jmno souboru, do kterho konzument zapisuje
     *              pedvan zznamy.
     *  @throws FileNotFoundException pokud soubor <i>_output_file_name</i>
     *              neexistuje a nepodailo se ho vytvoit
     *  @see #initialize
     */
    public TextStreamConsumer(String _output_file_name) throws FileNotFoundException {
        PrintWriter output_writer = new PrintWriter(_output_file_name);
        
        initialize(output_writer, true);
    }
    
}
