/*
 * TextFileProvider.java
 *
 * Created on 15. kvten 2005, 23:03
 */

package cz.matfyz.sykora.nngui.data;

import java.io.*;
import java.util.*;
import cz.matfyz.sykora.nngui.exception.*;

/**
 * Zkladn implementace zdroje dat. Hodnoty se nataj z textovho souboru,
 * jednotliv sla jsou oddlena mezerou, na dku je uloen vdy jeden zznam
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class TextStreamProvider implements DataProvider {
    
    /**
     * Data prv zpracovan dky, natan bhem <i>fetchRow</i>
     */
    private double[] currentLine;
    /**
     * Pokud si provider sm vytvel Reader pro pstup k souboru s daty, je
     * zde na nl uloen odkaz tak, aby ho mohl zavt po zavoln metody
     * <i>close</i>
     *  @see #close
     */
    private Reader inputStreamReader;
    /**
     * Oznauje, jestli byl zznam z vstupnho souboru pednaten bez dosti
     * uivatele (napklad pokud se dotazujeme na poet hodnot v zznamu ve
     * chvli kdy dn zznam jet nebyl naten
     *  @see #fetchRow
     */
    private boolean prefetchedRow;
    /**
     * Tokenizer zodpovdn za natn dat ze vstupnho proudu
     *  @see #initialize
     */
    private StreamTokenizer tokenizer;
    
    /**
     * Zave providera. Pokud provider zn Reader, ze kterho jsou tena data
     * tokenizerem, zave ho taky.
     *  @see #inputStreamReader
     */
    public void close() {
        if(inputStreamReader != null) {
            try {
                inputStreamReader.close();
            }
            catch(IOException io) {
                // pokud se nepodailo reader zavrit, tak uz to je fakt divny...
                // vypis chybu do std. vystupu a pokracuj
                System.err.println("TextStreamProvider.close(): failed to close " +
                                   "inputStreamReader");
            }
            inputStreamReader = null;
        }
    }
    
    /**
     * Natn dky ze vstupnho proudu. Pokud byla dka naten v rmci
     * zjiovn velikosti zznamu, vrac tu pednatenou. Data oekv v
     * textovm formtu, jednotliv hodnoty oddlen jednou, nebo vce mezerami,
     * na kadm dku je prv jeden zznam.
     *  @return true pokud byl ze vstupu spn naten dal zznam, jinak
     *              false pokud vstup neobsahuje dn dal zznamy
     *  @throws InvalidDataFormatException pokud data ve vstupnm proudu nejsou
     *              ve sprvnm formtu (vstup obsahuje jin hodnoty ne sla
     *              a mezery, nebo na nkterm dku nen sprvn poet hodnot.
     *  @see #currentLine
     *  @see #prefetchedRow
     *  @see #tokenizer
     */
    public boolean fetchRow() throws DataException {
        if(prefetchedRow) {
            prefetchedRow = false;
            return true;
        }
        
        try {
            if(currentLine == null) {
                Vector<Double> values = new Vector<Double>();

                tokenizer.nextToken();
                if(tokenizer.ttype == StreamTokenizer.TT_EOF)
                    return false;
                
                while(tokenizer.ttype != StreamTokenizer.TT_EOL) {
                    if(tokenizer.ttype != StreamTokenizer.TT_NUMBER)
                        throw new InvalidDataFormatException("Number expected");
                    
                    values.add(new Double(tokenizer.nval));
                    tokenizer.nextToken();
                }
                
                currentLine = new double[values.size()];
                for(int i = 0; i < values.size(); i++)
                    currentLine[i] = values.get(i).doubleValue();
            }
            else {
                tokenizer.nextToken();
                if(tokenizer.ttype == StreamTokenizer.TT_EOF)
                    return false;
                tokenizer.pushBack();
                
                for (int i = 0; i < currentLine.length; i++) {
                    tokenizer.nextToken();
                    if(tokenizer.ttype != StreamTokenizer.TT_NUMBER)
                        throw new InvalidDataFormatException("Number expected");
                    
                    currentLine[i] = tokenizer.nval;
                }
                // prebyvajici cisla nam nijak nevadi, preskoc je az na konec
                // radky
                while(tokenizer.ttype == StreamTokenizer.TT_NUMBER)
                    tokenizer.nextToken();
            }
        }
        catch(IOException io) {
            return false;
        }
        
        return true;
    }
    
    /**
     * Zjitn velikosti zznamu providera. Pokud provider jet nenaetl dn
     * zznam, pednate prvn, aby z nj mohl dopotat jeho velikost.
     *  @return poet hodnot v zznamu providera
     *  @see #prefetchedRow
     */
    public int getRowSize() {
        if(currentLine == null) {
            try {
                if(!fetchRow())
                    return 0;
                prefetchedRow = true;
            }
            catch(DataException e) {
                return 0;
            }
        }
        return currentLine.length;
    }
    
    /**
     * Zjitn hodnoty na zadanm indexu.
     *  @param _index index zjiovan hodnoty
     *  @return hodnota na indexu <i>_index</i>
     *  @throws DataProviderNotReadyException pokud do provideru dosud nebyl
     *              naten dn zznam.
     *  @see #currentLine
     */
    public double getValue(int _index) throws DataException {
        if(currentLine == null)
            throw new DataProviderNotReadyException(this);
        
        return currentLine[_index];
    }
    
    /**
     * Inicializace provideru, kter je spolen pro vechny konstruktory.
     * Vytvo tokenizer nad vstupnm readerem.
     *  @param _input reader, kter slou pro natn vstupnch dat
     */
    private void initialize(Reader _input) {
        assert(_input != null);
        
        prefetchedRow = false;
        
        tokenizer = new StreamTokenizer(_input);
        tokenizer.eolIsSignificant(true);
        tokenizer.parseNumbers();
    }
    
    /**
     * Petoen providera zpt na prvn zznam. Uspje pokud provider vlastn
     * referenci na vstupn reader a tento podporuje nastavovn pozice.
     *  @return true pokud se podailo pozici nastavit na zatek vstupnho
     *              proudu, false pokud se pozici v proudu zmnit nepodailo
     *  @see #inputStreamReader
     */
    public boolean rewind() {
        if(inputStreamReader != null) {
            try {
                inputStreamReader.reset();
            }
            catch (IOException io) {
                return false;
            }
            return true;
        }
        
        return false;
    }
    
    /**
     * Konstruktor. Inicializuje provider pro ten ze zadanho souboru.
     *  @param _input_file objekt reprezentujc soubor, ze kterho budou tena
     *              vstupn data
     *  @see #initialize
     */
    public TextStreamProvider(File _input_file) throws FileNotFoundException {
        if(_input_file == null)
            throw new NullPointerException("Input file must not be null");
        
        inputStreamReader = new FileReader(_input_file);
        initialize(inputStreamReader);
    }
    
    /**
     * Konstruktor. Inicializuje provider pro zadan reader.
     *  @param _input Reader, kter se m pout pro ten dat ze vstupnho
     *              proudu
     *  @see #initialize
     */
    public TextStreamProvider(Reader _input) {
        if(_input == null)
            throw new NullPointerException("Input stream reader must be specified!");
        
        inputStreamReader = null;
        initialize(_input);
    }
    
    /**
     * Konstruktor. Inicializuje provider pro ten ze zadanho souboru.
     *  @param _input_file_name jmno souboru, ze kterho se maj st data
     *  @throws FileNotFoundException pokud soubor <i>_input_file_name</i>
     *              neexistuje, nebo se ho nepodailo otevt pro ten
     *  @see #initialize
     */
    public TextStreamProvider(String _input_file_name) throws FileNotFoundException {
        if(_input_file_name == null)
            throw new NullPointerException("Input stream file name must not be null");
        
        inputStreamReader = new FileReader(_input_file_name);
        initialize(inputStreamReader);
    }
}
