/*
 * ArchitectureEditorDialog.java
 *
 * Created on 1. erven 2005, 20:39
 */

package cz.matfyz.sykora.nngui.gui;

import javax.swing.*;
import javax.swing.table.*;

/**
 * Dialog pro editaci architektury st.
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class ArchitectureEditorDialog extends JDialog {
    /**
     * Minimln architektura neuronov st, obsahuje jeden neuron, kter je
     * souasn vstupnm i vstupnm.
     */
    private final static int[] DEFAULT_ARCHITECTURE = { 1 };
    
    /**
     * Pomocn tda - model pro tabulku editujc architekturu st. Roziuje
     * DefaultTableModel o inicializaci ze zadan architektury.
     */
    private static class ArchitectureTableModel extends DefaultTableModel {
        /**
         * Pole obsahujc nzvy sloupc tabulky.
         *  @see #initializeFromArchitecture
         */
        private final static String[] COLUMN_LABELS = { "Layer", "Neuron count" };
        
        /**
         * Rozhodovn, jestli je zadan buka tabulky editovateln. Buka je
         * editovateln prv kdy obsahuje slo - velikost vrstvy
         *  @param _row dek, na kterm se buka nachz
         *  @param _col sloupec, ve kterm se buka nachz
         *  @return true pokud buka je editovateln, jinak false
         */
        public boolean isCellEditable(int _row, int _col) {
            // editovatelna jsou jenom policka s cisly - velikostmi vrstev
            return _col == 1;
        }
        
        /**
         * Rozhodovn o tom, jakho typu jsou data v zadanm sloupeku.
         *  @param _column_index index sloupeku
         *  @return <i>String.class</i> pro sloupek s popisky,
         *          <i>Integer.class</i> pro sloupeek s velikostmi vrstev
         */
        public Class getColumnClass(int _column_index) {
            switch(_column_index) {
                case 0:
                    return String.class;
                case 1:
                    return Integer.class;
                default:
                    assert(false);
                    return null;
            }
        }
        
        /**
         * Inicializuje tabulku podle zadan architektury, piprav popisky
         * sloupc a adek pro editaci.
         *  @param _architecture architektura st, pro jej velikost je s
         *              pipravovan
         */
        public void initializeFromArchitecture(int[] _architecture) {
            if(_architecture == null)
                throw new NullPointerException("_architecture must not be null");
            if(_architecture.length == 0)
                throw new IllegalArgumentException("_architecture array must not be empty");
            
            setColumnCount(2);
            setColumnIdentifiers(COLUMN_LABELS);
            setRowCount(_architecture.length);
            
            for(int i = 0; i < getRowCount(); i++) {
                setValueAt("Layer " + i, i, 0);
                setValueAt(new Integer(_architecture[i]),  i, 1);
            }
        }
        
        /**
         * Nastav hodnotu na zadan pozici v tabulce. Pokud zadvan hodnota je
         * velikost vrstvy, zkontroluje, jestli je vt, nebo rovn jedn.
         *  @param _value nastavovan hodnota
         *  @param _row dek do kterho je hodnota nastavovan
         *  @param _column sloupec, do kterho je hodnota nastavovan
         */
        public void setValueAt(Object _value, int _row, int _column) {
            if(_column == 1 && ((Number)_value).intValue() < 1)
                _value = new Integer(1);
            super.setValueAt(_value, _row, _column);
        }
        
        /**
         * Konstruktor. Inicializuje model pro zadanou architekturu
         *  @param _architecture architektura st, pro kterou m bt model
         *              inicializovn
         *  @see #initializeFromArchitecture
         */
        public ArchitectureTableModel(int[] _architecture) {
            initializeFromArchitecture(_architecture);
        }
    }
    
    /**
     * Konstruktor. Inicializuje model tabulky pro implicitn jednoneuronovou
     * architekturu.
     *  @param _parent rodiovsk okno pro tento dialog
     *  @see #initializeData
     */
    public ArchitectureEditorDialog(java.awt.Frame _parent) {
        super(_parent, true);
        initComponents();
        initializeData(DEFAULT_ARCHITECTURE);
    }
    
    /**
     * Konstruktor. Inicializuje model tabulky pro zadanou architekturu
     *  @param _parent rodiovsk okno pro tento dialog
     *  @param _architecture architektura st, pro kterou m bt dialog
     *              incializovn
     *  @see #initializeData
     */
    public ArchitectureEditorDialog(java.awt.Frame _parent, int[] _architecture) {
        super(_parent, true);
        initComponents();
        initializeData(_architecture);
    }
    
    /**
     * Export architektury st z tabulky do pole s velikostmi jednotlivch
     * vrstev. Toto pole je pak mon pout pi vytven st pomoc
     * <i>NetworkBuilder</i>.
     *  @return pole obsahujc architekturu st
     */
    public int[] exportData() {
        int[] architecture = new int[architectureTable.getRowCount()];
        for(int i = 0; i < architecture.length; i++)
            architecture[i] = ((Integer)architectureTable.getValueAt(i, 1)).intValue();
        return architecture;
    }
    
    /**
     * Inicializace ovldacch prvk formule. Automaticky generovno NetBeans
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        javax.swing.JScrollPane architectureScrollPane;
        javax.swing.JLabel numOfLayersLabel;

        numOfLayersLabel = new javax.swing.JLabel();
        rowCountEditor = new javax.swing.JSpinner();
        layerSizesLabel = new javax.swing.JLabel();
        architectureScrollPane = new javax.swing.JScrollPane();
        architectureTable = new javax.swing.JTable();

        getContentPane().setLayout(new javax.swing.BoxLayout(getContentPane(), javax.swing.BoxLayout.Y_AXIS));

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        setTitle("Edit architecture");
        numOfLayersLabel.setText("Numer of layers:");
        getContentPane().add(numOfLayersLabel);

        rowCountEditor.setMaximumSize(new java.awt.Dimension(32767, 20));
        rowCountEditor.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                rowCountEditorStateChanged(evt);
            }
        });

        getContentPane().add(rowCountEditor);

        layerSizesLabel.setText("Layer sizes:");
        getContentPane().add(layerSizesLabel);

        getContentPane().add(architectureTable.getTableHeader());
        architectureScrollPane.setViewportView(architectureTable);

        getContentPane().add(architectureScrollPane);

        pack();
    }
    // </editor-fold>//GEN-END:initComponents
    
    /**
     * Inicializace dat v ovldacch prvcch dialogu.
     *  @param _architecture architektura, pro kterou jsou ovldac prvky
     *              inicializovny
     *  @see ArchitectureEditorDialog.ArchitectureTableModel
     */
    protected void initializeData(int[] _architecture) {
        rowCountEditor.setValue(new Integer(_architecture.length));
        architectureTable.setModel(new ArchitectureTableModel(_architecture));
    }

    /**
     * Obsluha zmny v editoru potu vrstev neuronov st. Pid, nebo ubere
     * podle poteby rdky z tabulky pro jej editaci
     *  @param evt informace o udlosti GUI
     *  @see #rowCountEditor
     *  @see #initializeData
     */
    private void rowCountEditorStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_rowCountEditorStateChanged
        int current_rows = architectureTable.getRowCount();
        int new_rows = ((Integer)rowCountEditor.getValue()).intValue();
        
        if(new_rows < 1) {
            rowCountEditor.setValue(new Integer(1));
            return;
        }
        
        if(current_rows != new_rows) {
            int[] new_architecture = new int[new_rows];
            int i;
            for(i = 0; i < Math.min(current_rows, new_rows); i++)
                new_architecture[i] = ((Integer)architectureTable.getValueAt(i, 1)).intValue();
            for(;i < new_rows; i++)
                new_architecture[i] = 1;
            initializeData(new_architecture);
        }
    }//GEN-LAST:event_rowCountEditorStateChanged
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JTable architectureTable;
    private javax.swing.JLabel layerSizesLabel;
    private javax.swing.JSpinner rowCountEditor;
    // End of variables declaration//GEN-END:variables
    
}
