/*
 * Network.java
 *
 * Created on 4. kvten 2005, 15:35
 */

package cz.matfyz.sykora.nngui.network;

import cz.matfyz.sykora.nngui.exception.NoDataSourceException;
import java.util.*;
import cz.matfyz.sykora.nngui.data.*;
import cz.matfyz.sykora.nngui.exception.*;

/**
 * Zapouzden dopedn st jako celku, implementuje operace nad neuronovu st
 * jako nad celkem.
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class Network {
    
    /**
     * Genertor nhodnch sel, je spolen pro celou s
     *  @see #getSmallRandomNumber
     */
    private static Random randomGenerator = new Random(12345678);
    
    /**
     * Datov zdroj, ze kterho s erp vstupn data
     *  @see #fetchData
     */
    private DataProvider input;
    /**
     * Datov zdroj, ze kterho s erp poadovan vstupy ke vstupnm datm
     *  @see #fetchDesiredOutput
     */
    private DataProvider desiredOutput;
    /**
     * Model st pouit pi vpotu hodnot a pi jej adaptaci
     *  @see #calculateTransition
     */
    private NetworkModel model;
    /**
     * Konzument, kter ze st odebr vstupn hodnoty neuron z vstupn
     * (posledn) vrstvy
     */
    private DataConsumer output;
    /**
     * Seznam vrstev st, vstupn vrstva m index 0, vstupn size() - 1.
     * V kad sti mus bt alespo jedna vrstva
     */
    private Vector<NeuronRow> networkRows;
    
    /**
     * Adaptace na jednotlivch vrstvch st, provede pravu vah podle rozdl
     * v differenceTotal
     *  @param _strategy strategie podle kter se adaptace provd
     */
    protected void adapt(AdaptationStrategy _strategy) {
        assert(_strategy != null);
        
        for (int i = networkRows.size() - 1; i > 0; i--) {
            NeuronRow row = networkRows.get(i);
            row.adapt(_strategy);
        }
    }
    
    /**
     * Provede pravu vah st po zadanm potu trnovacch klad (dvkov
     * zpracovn trnovacch pklad)
     *  @param _strategy strategie podle kter se provd adaptace vah
     *  @param _iterations poet zpracovanch pklad ped pravou vah
     */
    public long adapt(AdaptationStrategy _strategy, long _iterations) throws DataException {
        long count = 0;
        if(_strategy == null)
            throw new NullPointerException("_strategy must not be null");
  
        while(_iterations-- > 0) {
            resetDifferences();
            
            if(!calculateSingle())
                break;
            if(!fetchDesiredOutput())
                break;
            
            count++;
            processDifferences();
        }
        adapt(_strategy);
        
        return count;
    }
    
    /**
     * Adaptace po jednom kroku vpotu. Chyba en v sti se nekumuluje,
     * ale hned po vpotu dojde k jejmu zpracovn.
     *  @param _strategy adaptan strategie pouit pi prav vah st
     *  @return true pokud adaptace probhla v podku, jinak false
     */
    public boolean adaptSingle(AdaptationStrategy _strategy) throws DataException {
        if(_strategy == null)
            throw new NullPointerException("_strategy must not be null!");
        
        resetDifferences();
        
        if(!calculateSingle())
            return false;
        if(!fetchDesiredOutput())
            return false;
        
        processDifferences();
        adapt(_strategy);
        
        return true;
    }
    
    /**
     * Pidn vrstvy neuron do st za posledn vrstvu (aktuln vstupn) a
     * jejich propojen s existujcmi neurony v sti
     *  @param _row nov pidvan vrstva neuron
     */
    protected void appendRow(NeuronRow _row) {
        if(_row == null)
            throw new IllegalArgumentException("_row must not be null!");
        
        if(networkRows.size() > 0) {
            NeuronRow last_row = networkRows.lastElement();
            _row.initializeConnections(last_row);
        }
        networkRows.add(_row);
    }
    
    /**
     * Vpoet neuronov st pro nejve <i>_iterations</i> zadn.
     *  @param _iterations poet zznam, kter se maj zpracovat
     *  @return poet zznam, kter byly skuten zpracovan
     */
    public long calculate(long _iterations) throws DataException {
        long calculated = 0;
        while(_iterations-- > 0) {
            if(!calculateSingle())
                break;
            calculated++;
        }
        
        return calculated;
    }
    
    /**
     * Vpoet neuronov st pro vechna dostupn zadn. Pracuje dokud jsou
     * njak zadn na vstupu a dokud je vstup ochoten pijmat data.
     *  @return poet zznam pro kter byl proveden vpoet
     *  @throws DataException pokud dojde k njakm problmm na producentovi,
     *      nebo na konzumentech dat
     */
    public long calculateAll() throws DataException {
        long iterations = 0;
        while(calculateSingle())
            iterations++;
        
        return iterations;
    }
    
    /**
     * Vpoet neuronov st pro jeden zznam. Provede se vetn naten dat
     * a jejich nslednho zpracovn do vstupu
     *  @return true pokud vpoet probhl bez problm, false pokud provider,
     *      nebo konzument dat u nejsou schopni pijmat dal data
     *  @throws DataException pokud dojde k njakm problmm na producentovi,
     *      nebo na konzumentech dat
     */
    public boolean calculateSingle() throws DataException {
        if(!fetchData())
            return false;
        
        for(int i = 1; i < networkRows.size(); i++)
            networkRows.get(i).calculate();
        
        return commitData();
    }
    
    /**
     * Vpoet pechodov funkce pro zadanou excitaci neuronu
     *  @param _excitation excitace neuronu (souet vench vstup a prahu)
     *  @return hodnota pechodov funkce pro zadanou excitaci
     *  @see #model
     *  @see NetworkModel#transitionFunction
     */
    protected final double calculateTransition(double _excitation) {
        return model.transitionFunction(_excitation);
    }
    
    /**
     * Vpoet derivace pechodov funkce pro zadanou excitaci neuronu
     *  @param _excitation excitace neuronu (souet vench vstup a prahu)
     *  @return hodnota derivace pechodov funkce pro zadanou excitaci
     *  @see #model
     *  @see NetworkModel#transitionFunctionDerivation
     */
    protected final double calculateTransitionDerivation(double _excitation) {
        return model.transitionFunctionDerivation(_excitation);
    }
    
    /**
     * Ped data z vstupn vrstvy st konzumentovi
     *  @return true pokud konzument pijal pedan data, false pokud konzumenta
     *              dal data u nezajmaj a nechce je pijmat
     *  @throws DataException pokud dojde k problmm u konzumenta
     */
    protected boolean commitData() throws DataException {
        assert(networkRows != null && networkRows.size() > 0);
        
        // pokud nikdo neposloucha na vystupu, tak ten vystup nejspis nikoho
        // nezajima. Takze se s tim nebudeme patlat a zahodime ho
        if(output == null)
            return true;
        
        NeuronRow row = networkRows.lastElement();
        for(int i = 0; i < row.getSize(); i++)
            output.setValue(i, row.getNeuron(i).getOutput());
        
        return output.commitRow();
    }
    
    /**
     * Zpracovn selnch dat jednoho zznamu ze vstupu a jejich naten jako
     * vstupn hodnoty do vstupn vrstvy neuron. Prost to nate data ze
     * vstupu a str je na zatek vstupu :)
     *  @return true pokud je s pipravena pro vpoet, false, pokud ve
     *              vstupnm souboru u nejsou dn data
     *  @throws DataException pokud pi natn dat dojde k njak chyb
     */
    protected boolean fetchData() throws DataException {
        assert(networkRows != null && networkRows.size() > 0);
        
        if(input == null)
            throw new NoDataSourceException(this);
        
        if(!input.fetchRow())
            return false;
        
        NeuronRow row = networkRows.firstElement();
        for(int i = 0; i < row.getSize(); i++)
            row.getNeuron(i).setOutput(input.getValue(i));
        
        return true;
    }
    
    /**
     * Zpracovn selnch dat poadovanho vstupu k aktuln zpracovvanmu
     * zznamu ze vstupnho souboru. Pedpokld, e data ze vstupnho souboru
     * byla ji natena
     *  @return true pokud se oekvan vstupy podailo v podku nast,
     *              false pokud soubor oekvanch vstup neobsahuje dn data
     *  @throws DataException pokud pi natn dat dojde k njak chyb
     */
    protected boolean fetchDesiredOutput() throws DataException {
        assert(networkRows != null && networkRows.size() > 0);
        
        if(desiredOutput == null)
            throw new NoDataSourceException(this);
        
        if(!desiredOutput.fetchRow()) {
            // cosi prohnileho je ve state Danskem - (zrejme) se podarilo nacist
            // data ze vstupniho souboru, ale pozadovany vystup uz se k nim
            // nenacetl; to by se stat nemelo!
            assert(false);
            return false;
        }
        
        NeuronRow row = networkRows.get(networkRows.size() - 1);
        for (int i = 0; i < row.getSize(); i++) {
            // nastav rozdil mezi pozadovanym a skutecnym vystupem pro neurony
            // ve vystupni vrstve
            row.getNeuron(i).addDifference(desiredOutput.getValue(i) - row.getNeuron(i).getOutput());
        }
        
        return true;
    }
    
    /**
     * Dotaz na architekturu st (poet vrstev a jejich velikosti). Vrac
     * informace o architektue ve form pole integer, kad prvek pole
     * reprezentuje jednu vrstvu, vstupn vrsvta m index 0, vstupn vrstva m
     * index <i>networkRows.size() - 1</i>
     *  @return pole s informacemi o architektue st
     *  @see NetworkBuilder#createNetwork
     */
    public int[] getArchitecture() {
        assert(networkRows != null && networkRows.size() > 0);
        
        int[] architecture = new int[networkRows.size()];
        for(int i = 0; i < networkRows.size(); i++)
            architecture[i] = networkRows.get(i).getSize();
        
        return architecture;
    }
    
    /**
     * Dotaz na velikost (poet neuron) vstupn vrstvy st.
     *  @return poet neuron ve vstupn vrstv st
     */
    public int getInputRowSize() {
        assert(networkRows != null && networkRows.size() > 0);
        
        return networkRows.get(0).getSize();
    }
    
    /**
     * Dotaz na model st (objekt, kter realizuje vpoet pechodov funkce a
     * jej prvn derivace).
     *  @return odkaz na objekt reprezentujc model tto st
     */
    public final NetworkModel getModel() {
        assert(model != null);
        
        return model;
    }
    
    /**
     * Pstup k jednotlivm neuronm v sti. Pozor, zmna na neuronu zstane
     * v sti zachovna a me znehodnotit vsledek uen.
     *  @param _row vrstva st, ve kter se neuron nachz. Vstupn vrstva m
     *              index 0, vstupn m index getRowCount() - 1
     *  @param _index index neuronu ve vrstv st
     *  @return odkaz na objekt reprezentujc neuron na zadan pozici v sti
     *  @throws IndexOutOfBounds pokud <i>_row</i> a <i>_index</i> nejsou platn
     *              souadnice neuronu v sti
     */
    public Neuron getNeuron(int _row, int _index) {
        assert(networkRows != null);    
        if(_row < 0 || _row >= getRowCount())
            throw new IndexOutOfBoundsException("_row must be between 0 and getRowCount() (" + getRowCount() + ")!");
        
        NeuronRow row = networkRows.get(_row);
        assert(row != null);
        
        if(_index < 0 || _index >= row.getSize())
            throw new IndexOutOfBoundsException("_index must be between 0 and row.getSize() (" + row.getSize() + ")");
        
        return row.getNeuron(_index);
    }
    
    /**
     * Dotaz na velikost vstupn vrstvy st.
     *  @return poet neuron ve vstupn vrstv st.
     */
    public int getOutputRowSize() {
        assert(networkRows != null && getRowCount() > 0);
        
        return networkRows.get(getRowCount() - 1).getSize();
    }
    
    /**
     * Pstup k jednotlivm vrastvm neuron v sti. Pozor, zmna na vrstv
     * neuron zstane v sti zachovna a me znehodnotit vsledky uen.
     *  @param _row index vrstvy st. Vstupn vrstva m index 0, smrem k
     *              vstupn vrstv indexy rostou.
     *  @return odkaz na objekt reprezentujc vrstvu st s indexem <i>_row</i>
     */
    public NeuronRow getRow(int _row) {
        assert(networkRows != null);
        if(_row < 0 || _row >= getRowCount())
            throw new IndexOutOfBoundsException("_row must be between 0 and getRowCount() (" + getRowCount() + ")!");
        
        return networkRows.get(_row);
    }
    
    /**
     * Zjitn potu vrstev st. Vstupn i vstupn vrstva se do velikosti
     * st potaj
     *  @return poet vrstev neuronov st
     */
    public final int getRowCount() {
        assert(networkRows != null);
        
        return networkRows.size();
    }
    
    /**
     * Genertor malch nhodnch sel s gaussovskm rozloenm s stedn
     * hodnotou 0 a rozptylem 1
     *  @return mal nhodn lo
     *  @see #randomGenerator
     */
    protected final double getSmallRandomNumber() {
        return randomGenerator.nextGaussian();
    }
    
    /**
     * Inicializace spojen mezi neurony sousednch vrstev a jejich nastaven
     * na mal nhodn hodnty.
     */
    protected void initializeAllConnections() {
        for (int i = 1; i < networkRows.size(); i++) {
            NeuronRow row = networkRows.get(i);
            NeuronRow prev = networkRows.get(i - 1);
            
            row.initializeConnections(prev);
        }
    }

    /**
     * Zptn propagace chyby v sti podle algoritmu backpropagation. Provede
     * zptn en postupn na vech vrstvch od vstupn smrem ke vstupn.
     * Na rovni st se pouze pedv zen jednotlivm vrstvm ve sprvnm
     * poad.
     *  @see NeuronRow#processDifferences
     */
    protected void processDifferences() {
        assert(networkRows != null && networkRows.size() > 0);
        
        for (int i = networkRows.size() - 1; i > 0; i--)
            networkRows.get(i).processDifferences();
    }
    
    /**
     * Reset chyby aktulnho zadn na vech neuronech zpt na 0
     *  @see #adapt(AdaptationStrategy, long)
     *  @see #adaptSingle
     */
    protected void resetDifferences() {
        assert(networkRows != null && networkRows.size() > 0);
        
        for (int i = 0; i < networkRows.size(); i++)
            networkRows.get(i).resetDifferences();
    }
    
    /**
     * Nastav providera, ze kterho s erp poadovan vstup pro vzorky
     * zpracovvan bhem jejho uen. Oekv se, e poadovan vstup naten
     * pomoc <i>desiredOutput.fetchRow()</i> odpovd poslednmu zadn
     * natenmu pomoc <i>input.fetchRow()</i> ped volnm
     * <i>desiredOutput.fetchRow()</i>.
     * Pokud s u m vybudovanou strukturu, mus velikost zznamu providera
     * odpovdat velikosti vstupn vrstvy st.
     *  @param _provider provider, kter m slouit jako zdroj oekvanch
     *              vstup pro zpracovvan vstupn vzory
     *  @throws NullPointerException pokud <i>_provider</i> je null
     *  @throws RecordSizesDoNotMatchException pokud velikost zznamu providera
     *              neodpovd potu neuron ve vstupn vrstv
     *  @see #desiredOutput
     *  @see #input
     *  @see #setInputProvider
     */
    public void setDesiredOutputProvider(DataProvider _provider) throws DataException {
        // TODO: tady by to mozna slo i bez providera..
        if(_provider == null)
            throw new NullPointerException("Provider must not be null");
        
        if(getRowCount() > 0) {
            int new_output_row_size = _provider.getRowSize();
            int old_output_row_size = networkRows.get(getRowCount() - 1).getSize();
            
            if(new_output_row_size != old_output_row_size)
                throw new RecordSizesDoNotMatchException("New desired output provider " +
                                "record size does not match input row size");
        }
        
        desiredOutput = _provider;
    }
    
    /**
     * Nastav providera, ze kterho s nat zadn pro vpoet, nebo pro
     * uen. V takovm ppad se oekv, e tento provider bude
     * synchronizovn s providerem poskytujcm oekvan vstupy st, jak je
     * popsno v dokumentaci k <i>setDesiredOutputProvider</i>.
     * Pokud s u m vybudovanou strukturu, mus velikost zznamu providera
     * odpovdat potu neuron ve vstupn vrstv st.
     *  @param _provider provider, kter m slouit jako zdroj vstupnch dat
     *              pro vpoty, nebo uen st
     *  @throws NullPointerException pokud <i>_provider</i> je null
     *  @throws RecordSizesDoNotMatchException pokud velikost zznamu providera
     *              neodpovd potu neuron ve vstupn vrstv
     *  @see #desiredOutput
     *  @see #output
     *  @see #setDesiredOutputProvider
     */
    public void setInputProvider(DataProvider _provider) throws DataException {
        if(_provider == null) {
            // throw new NullPointerException("Provider must not be null");
            input = _provider;
            return;
        }
        
        if(getRowCount() > 0) {
            int new_input_row_size = _provider.getRowSize();
            int old_input_row_size = networkRows.get(0).getSize();
            
            if(new_input_row_size != old_input_row_size)
                throw new RecordSizesDoNotMatchException("New data provider record size " +
                                "does not match input row size");
        }
        
        input = _provider;
    }
    
    /**
     * Nastav konzumenta, do kterho bude s bhem vpotu a uen ukldat sv
     * vstupn hodnoty.
     * Pokud s u m vybudovanou strukturu, mus bt velikost vstupn vrstvy
     * st pro konzumenta akceptovateln, tj. voln
     * <i>_consumer.setRowSize</i> pro velikost vstupn vrstvy st mus vracet
     * true.
     *  @param _consumer konzument, do kterho bude s ukldat sv vstupn
     *              hodnoty
     *  @see #output
     *  @see DataConsumer#setRowSize
     *  @throws RecordSizesDoNotMatchException pokud velikost vstupn vrstvy
     *              st nen pro konzumenta akceptovateln (ve smyslu
     *              <i>DataConsumer.setRowSize</i>)
     */
    public void setOutputConsumer(DataConsumer _consumer) throws DataException {
        if(_consumer == null) {
            // konzument muze byt null - sit nikam nic neposila a ma to dobry smysl...
            // takze ten throw dobry smysl nema
            // throw new NullPointerException("Consumer must not be null");
            output = _consumer;
            return;
        }
        
        if(getRowCount() > 0) {
            int new_output_row_size = _consumer.getRowSize();
            int old_output_row_size = networkRows.get(getRowCount() - 1).getSize();
            
            if(new_output_row_size != old_output_row_size
                    && !_consumer.setRowSize(old_output_row_size))
                throw new RecordSizesDoNotMatchException("New output consumer record " +
                                "size does not match output row size");
        }
        
        output = _consumer;
    }
    
    /**
     * Konstruktor. Vytv novou instanci st pro zadan model.
     *  @param _model model st, kter m bt pouit
     *  @throws NullPointerException pokud je zadan reference na model nulov
     */
    protected Network(NetworkModel _model) {
        model = _model;
        if(model == null)
            throw new NullPointerException("_model must not be null!");
        networkRows = new Vector<NeuronRow>();
    }    
}
