/*
 * NetworkBuilder.java
 *
 * Created on 22. kvten 2005, 13:34
 */

package cz.matfyz.sykora.nngui.network;

import cz.matfyz.sykora.nngui.exception.*;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.io.StringReader;
import org.w3c.dom.*;

/**
 * Objekt pro konstrukce neuronovch st podle zadanch parametr
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class NetworkBuilder {
    /**
     * Jmno XML elementu obsahujcho informace o architektue.
     *  @see #parseArchitecture
     */
    public static final String  ARCHITECTURE_NODE_NAME          = "architecture";
    /**
     * Jmno XML elementu obsahujcho informace o modelu.
     *  @see #parseModel
     */
    public static final String  MODEL_NODE_NAME                 = "model";
    /**
     * Jmno atributu obsahujc nzev modelu st (podle tohoto nzu je model
     * vytven pi natn st z XML souboru.
     *  @see #parseModel
     */
    public static final String  MODEL_NAME_ATTRIBUTE_NAME       = "name";
    /**
     * Jmno XML elementu obsahujcho jednotliv vhy pro jeden neuron
     *  @see #parseWeights
     */
    public static final String  NEURON_WEIGHTS_NODE_NAME        = "neuron";
    /**
     * Jmno XML elementu obsahujcho vechny informace o neuronov sti jako
     * sv synovsk uzly.
     *  @see #createNetwork(org.w3c.dom.Node)
     */
    public static final String  NETWORK_NODE_NAME               = "network";
    /**
     * Jmno atributu kdujcho typ st. Uruje zpsob zpsob s jakm se se
     * st dl nakld. (podporovny jsou pouze st typu back-propagation)
     *  @see #createNetwork(org.w3c.dom.Node)
     */
    public static final String  NETWORK_TYPE_ATTRIBUTE_NAME     = "type";
    /**
     * Jmno XML elementu kdujcho informace o velikosti jedn vrstvy st
     *  @see #parseArchitecture
     */
    public static final String  ROW_SIZE_NODE_NAME              = "row";
    /**
     * Jmno atributu ukldajcho velikost jedn vrstvy st
     *  @see #parseArchitecture
     */
    public static final String  ROW_SIZE_ATTRIBUTE_NAME         = "size";
    /**
     * Jmno XML elementu obsahujcho informace o vahch neuron jedn vrstvy
     * st.
     *  @see #parseWeights
     */
    public static final String  ROW_WEIGHTS_NODE_NAME           = "row";
    /**
     * Jmno atributu ukldajcho prh neuronu
     *  @see #parseWeights
     */
    public static final String  THRESHOLD_ATTRIBUTE_NAME        = "threshold";
    /**
     * Jmno XML elementu obsahujcho informace o vahch neuron v sti
     *  @see #parseWeights
     */
    public static final String  WEIGHTS_NODE_NAME               = "weights";
    
    /**
     * Tvorba nov dopedn neuronov st s danou architekturou (zadan poet
     * vrstev a poty neuron v jednotlivch vrstvch)
     *  @param _architecture popis architektury st. Nov vytvoen s m
     *              stejn poet vrstev jako je dlka pole, v kad vrstv je
     *              prv <i>_architecture[i]</i> neuron. Vstupn vrstva m
     *              index 0, vstupn vrstva m index _architecture.length - 1
     *  @param _model model, kter se m pro vytvenou s pout
     *  @return nov vytvoen neuronov s
     */
    public Network createNetwork(int[] _architecture, NetworkModel _model) {
        if(_model == null)
            throw new NullPointerException("_model must not be null");
        if(_architecture == null)
            throw new NullPointerException("_architecture must not be null");
        if(_architecture.length <= 0)
            throw new IllegalArgumentException("_architecture array must not be empty");
        
        Network net = new Network(_model);
        for(int i = 0; i < _architecture.length; i++) {
            if(_architecture[i] <= 0)
                throw new IllegalArgumentException("There must be at least one " +
                                                   "neuron in each row!");
            NeuronRow row = createRow(net, _architecture[i]);
            net.appendRow(row);
        }
        
        return net;
    }
    
    /**
     * Nahraje informace o struktue st z XML souboru (zpracovanho
     * pomoc XML DOM)
     *  @param _xml koen XML DOM stromu obsahujcho formt st
     *  @return nov vytvoenou s podle parametr v XML dokumentu
     *  @throws InvalidDataFormatException pokud struktura nebo obsah XML
     *              dokumentu neodpovdaj formtu pouitmu pro ukldn dat
     *              neuronov st
     *  @throws UnsupportedNetworkTypeException pokud je zadn typ neuronov
     *              st kter nen podporovn
     *  @throws UnsupportedNetworkModelException pokud je zadn model st,
     *              kter nen podporovn
     */
    public Network createNetwork(org.w3c.dom.Node _xml) throws InvalidDataFormatException,
                                                               UnsupportedNetworkTypeException,
                                                               UnsupportedNetworkModelException {
        if(null == _xml)
            throw new NullPointerException("_xml node must not be null");
        if(_xml.getNodeType() != Node.ELEMENT_NODE
                || !_xml.getNodeName().equals(NETWORK_NODE_NAME))
            throw new IllegalArgumentException("root node of the XML network representation must be a XML element named 'network'");
        
        Element root_element = (Element)_xml;
        
        // zkontroluj, ze je zadany typ site
        if(!root_element.hasAttribute(NETWORK_TYPE_ATTRIBUTE_NAME)) {
            System.err.println("Warning: NetworkBuilder.createNetwork(org.w3c.dom.Node): There is no " +
                    "network type specification attribute, defaulting to feed-forward back-propagation network");
        }
        else {
            // zkontroluj, ze sit je typu back-propagation, pokud ne, neumime
            // si s tim poradit a koncime
            String network_type = root_element.getAttribute(NETWORK_TYPE_ATTRIBUTE_NAME).trim();
            if(!network_type.equals("back-propagation"))
                throw new UnsupportedNetworkTypeException("Bad network type \'" + network_type + "\'");
        }
        
        // zpracuj architekturu site
        NodeList architecture_nodes = root_element.getElementsByTagName(ARCHITECTURE_NODE_NAME);
        
        assert(architecture_nodes != null);
        if(architecture_nodes.getLength() != 1)
            throw new InvalidDataFormatException("There must be exactly one architecture node");
        
        Node architecture_nd = architecture_nodes.item(0);
        assert(architecture_nd != null);
        assert(architecture_nd.getNodeType() == Node.ELEMENT_NODE);
        
        int[] architecture = parseArchitecture(architecture_nd);
        
        // vypreparuj informace o modelu site
        NetworkModel model = null;
        
        NodeList model_nodes = root_element.getElementsByTagName(MODEL_NODE_NAME);
        assert(model_nodes != null);
        
        int model_nodes_count = model_nodes.getLength();
        if(model_nodes_count == 1) {
            Node model_nd = model_nodes.item(0);
            assert(model_nd != null);
            
            model = parseModel(model_nd);
        }
        else if(model_nodes_count == 0) {
            // pokud neni zadan model site, predpokladej, ze sit pouziva 
            // sigmoidu pro prenos dat mezi neurony
            System.err.println("NetworkBuilder.createNetwork(org.w3c.dom.Node): " +
                               "no <model> tag found, defaulting to sigmoid model");
            
            model = new SigmoidModel();
        }
        else
            throw new InvalidDataFormatException("There must be at mose one model node");
        
        Network network = createNetwork(architecture, model);
        
        // pokud soubor obsahuje vahy, zkus je zpracovat a priradit neuronum
        NodeList weights_nodes = root_element.getElementsByTagName(WEIGHTS_NODE_NAME);
        if(weights_nodes != null && weights_nodes.getLength() > 0)
            parseWeights(weights_nodes.item(0), network);
        
        return network;
    }
    
    /**
     * Vytvoen vrstvy neuron dan velikosti pro dopenou s
     *  @param _network s pro kterou vrstvu neuron vytvme
     *  @param _neurons poet neuron v nov vytven vrstv
     *  @return nov vytvoen objekt reprezentujc vrstvu st
     */
    protected NeuronRow createRow(Network _network,int _neurons) {
        assert(_network != null && _neurons > 0);
        
        return new NeuronRow(_network,_neurons);
    }
    
    /**
     * Export informac o architektue st do XML reprezentace. Vytv
     * strukturu &lt;architecture&gt; v XML strom pro reprezentaci neuronov
     * st.
     *  @param _xml_doc XML dokument, do kterho se architektura ukld (pouito
     *              pro tvorbu novch XML element)
     *  @param _network s, jej architektura je ukldna
     *  @return nov XML element s reprezentac architektury st
     *  @throws DOMException pokud dojde k chyb pi prci s DOM stromem
     *  @see #exportNetworkToXML
     */
    protected Element exportArchitecture(org.w3c.dom.Document _xml_doc, Network _network) throws DOMException {
        assert(_xml_doc != null && _network != null);
        
        // vypreparuj ze site informace o architekture
        int[] architecture = _network.getArchitecture();
        
        // priprav XML reprezentaci
        Element architecture_root = _xml_doc.createElement(ARCHITECTURE_NODE_NAME);
        for(int i = 0; i < _network.getRowCount(); i++) {
            Element current_row_element = _xml_doc.createElement(ROW_SIZE_NODE_NAME);
            current_row_element.setAttribute(ROW_SIZE_ATTRIBUTE_NAME, Integer.toString(architecture[i]));
            
            architecture_root.appendChild(current_row_element);
        }
        
        return architecture_root;
    }
    
    /**
     * Export informac o modelu st do XML reprezentace. Vytv strukturu
     * &lt;model&gt; v XML strom pro reprezentaci neuronov st. Tato metoda
     * ukld pouze jmno modelu, uloen jeho konkrtnch parametr u nechv
     * na samotnm modelu a jeho metod NetworkModel#exportModelParameters.
     *  @param _xml_doc XML dokument, do kterho se informace o modelu ukldaj
     *              (pouito pro tvorbu novch XML element)
     *  @param _network s, jej model je ukldn
     *  @return nov XML element s reprezentac modelu st a nastavm jeho
     *              parametr
     *  @throws DOMException pokud dojde k chyb pi prci s DOM stromem
     *  @see #exportNetworkToXML
     *  @see NetworkModel#exportModelParameters
     */
    protected Element exportModel(org.w3c.dom.Document _xml_doc, Network _network) throws DOMException {
        assert(_xml_doc != null && _network != null);
        
        NetworkModel model = _network.getModel();
        Element model_root = _xml_doc.createElement(MODEL_NODE_NAME);
        
        model_root.setAttribute(MODEL_NAME_ATTRIBUTE_NAME, model.getModelName());
        model.exportModelParameters(model_root);
        
        return model_root;
    }
    
    /**
     * Export neuronov st do XML reprezentace. Vytv komletn strukturu
     * &lt;network&gt; s informacemi o sti.
     *  @param _xml_doc XML dokument, do kterho se informace o sti ukldaj
     *              (pouito pro tvorbu novch element)
     *  @param _network ukldn neuronov s
     *  @return nov XML element s reprezentac neuronov st
     *  @throws DOMException pokud dojde k chyb pi pri s DOM stromem
     *  @see #exportArchitecture
     *  @see #exportModel
     *  @see #exportWeights
     */
    public Node exportNetworkToXML(org.w3c.dom.Document _xml_doc, Network _network) throws DOMException {
        if(_network == null)
            throw new NullPointerException("_network must not be null");
        if(_xml_doc == null)
            throw new NullPointerException("_xml_doc must not be null");
        
        Element root = _xml_doc.createElement(NETWORK_NODE_NAME);
        root.setAttribute(NETWORK_TYPE_ATTRIBUTE_NAME, "back-propagation");
        
        Element architecture = exportArchitecture(_xml_doc, _network);
        root.appendChild(architecture);
        
        Element model = exportModel(_xml_doc, _network);
        root.appendChild(model);
        
        Element weights = exportWeights(_xml_doc, _network);
        root.appendChild(weights);
        
        return root;
    }
    
    /**
     * Export vah v neuronov sti. Vhy jsou exportovny pro vechny neurony,
     * v XML reprezentaci jsou uloeny postupn od vstupn vrstvy k vstupn,
     * jednotliv neurony jsou seazeny podle index od 0 dle, stejn jsou
     * seazeny ukldan vhy.
     *  @param _xml_doc XML dokuemtn, do kterho se informace o vahch ukldaj
     *              (pouito pro tvorbu novch element)
     *  @param _network s, jej vhy jsou ukldny
     *  @return nov XML element s reprezetac neuronov st
     *  @throws DOMException pokud dojde k chyb pi prci s DOM stromem
     *  @see #exportNetworkToXML
     */
    protected Element exportWeights(org.w3c.dom.Document _xml_doc, Network _network) throws DOMException {
        assert(_xml_doc != null && _network != null);
        
        Element weights_root = _xml_doc.createElement(WEIGHTS_NODE_NAME);
        for(int row = 1; row < _network.getRowCount(); row++) {
            NeuronRow currentRow = _network.getRow(row);
            Element row_weights = _xml_doc.createElement(ROW_WEIGHTS_NODE_NAME);
            
            for(int neuron = 0; neuron < currentRow.getSize(); neuron++) {
                Element neuron_weights = _xml_doc.createElement(NEURON_WEIGHTS_NODE_NAME);
                
                Neuron currentNeuron = currentRow.getNeuron(neuron);
                StringBuilder buffer = new StringBuilder();
                
                neuron_weights.setAttribute(THRESHOLD_ATTRIBUTE_NAME, Double.toString(currentNeuron.getThreshold()));
                
                double[] weigths = currentNeuron.getWeights();
                for(int weight = 0; weight < weigths.length; weight++) {
                    if(weight > 0)
                        buffer.append(' ');
                    buffer.append(weigths[weight]);
                }
                
                Text weights_data_node = _xml_doc.createTextNode(buffer.toString());
                neuron_weights.appendChild(weights_data_node);
                
                row_weights.appendChild(neuron_weights);
            }
            weights_root.appendChild(row_weights);
        }
        
        return weights_root;
    }
    
    /**
     * Zpracuje st XML DOM stromu, ve kter jsou uloeny informace
     * o architektue st (tedy o potu vrstev a jejich velikostech).
     *  @param _arch_node XML DOM element, ve kterm jsou uloeny data
     *              o jednotlivch vrstvch st
     *  @throws InvalidDataException pi chybn struktue XML DOM stromu,
     *              nebo pokud velikosti vrstev nejsou sprvn formtovan
     *              sla
     */
    protected int[] parseArchitecture(Node _arch_node) throws InvalidDataFormatException {
        assert(_arch_node != null);
        assert(_arch_node.getNodeType() == Node.ELEMENT_NODE);
        
        Element arch_elm = (Element)_arch_node;
        
        NodeList row_size_nodes = arch_elm.getElementsByTagName(ROW_SIZE_NODE_NAME);
        assert(row_size_nodes != null);
        
        int row_count = row_size_nodes.getLength();
        if(row_count < 1)
            throw new InvalidDataFormatException("There must be at least one " +
                                                 "row in the network");
        
        int[] architecture = new int[row_count];
        for(int i = 0; i < row_count; i++) {
            Node row_node = row_size_nodes.item(i);
            assert(row_node != null);
            assert(row_node.getNodeType() == Node.ELEMENT_NODE);
            assert(row_node.getNodeName().equals(ROW_SIZE_NODE_NAME));
            
            Element row_size = (Element)row_node;
            
            if(!row_size.hasAttribute(ROW_SIZE_ATTRIBUTE_NAME))
                throw new InvalidDataFormatException("There is no \"size\" attribute " +
                                                     "specified for a row");
            
            String size_string = row_size.getAttribute(ROW_SIZE_ATTRIBUTE_NAME);
            try {
                architecture[i] = Integer.parseInt(size_string);
            }
            catch(NumberFormatException num_format) {
                throw new InvalidDataFormatException("Size of a row is not valid " +
                                                     "number (" + size_string + ")", num_format);                        
            }
        }
        
        return architecture;
    }
    
    /**
     * Zpracuje XML reprezentaci modelu st. Nat typ modelu, vytv jeho
     * novou instanci a pedv ji element, aby v nm sama zpracovala sv
     * dal parametry.
     *  @param _model_node XML element s informacemi o modelu st
     *  @return novou instanci modelu st na zklad parametr z
     *              <i>_model_node</i>
     *  @throws InvalidDataFormatException pokud struktura, nebo obsah XML dat
     *              neodpovd formtu pouitmu pro uloen neuronov st
     *  @throws UnsupportedNetworkModelExcetion pokud je poedovn model st,
     *              kter nen znm, nebo nen podporovan
     */
    protected NetworkModel parseModel(org.w3c.dom.Node _model_node) throws InvalidDataFormatException,
                                                                           UnsupportedNetworkModelException {
        assert(_model_node != null);
        assert(_model_node.getNodeType() == Node.ELEMENT_NODE);
        
        Element model_elm = (Element)_model_node;
        
        // zjisti jmeno modelu z prislusneho atributu
        if(!model_elm.hasAttribute(MODEL_NAME_ATTRIBUTE_NAME))
            throw new InvalidDataFormatException("Model name was not specified " +
                                                 "although there was a <model> tag");
        String name = model_elm.getAttribute(MODEL_NAME_ATTRIBUTE_NAME);
        
        // pokud podle jmena dokazes najit pozadovany model, vytvor jeho instanci...
        NetworkModel model = null;
        if(name.equals(SigmoidModel.MODEL_NAME)) 
            model = new SigmoidModel();
        else
            // v opacnem pripade se jedna o chybu -> ohlasit a konec
            throw new UnsupportedNetworkModelException("Unknown network model \'" + name + "\'");
        
        // nacti parametry pro konkretni instanci modelu site
        model.loadModelParameters(model_elm);
        
        return model;
    }
    
    /**
     * Zpracuje XML reprezentaci vah st. Nat vhy pro jednotliv
     * vrstvy/neurony. Ovuje, e struktura pedvanch vah odpovd (dve
     * zpracovan) architektue st.
     *  @param _weights XML element s informacemi o vahch st
     *  @param _network s, pro kterou jsou v <i>_weights</i> uloeny vhy
     *  @throws InvalidDataFormatException pokud formt a obsah dat v XML DOM
     *              strom neodpovd formtu dat pro uloen vah st.
     */
    protected void parseWeights(org.w3c.dom.Node _weights, Network _network) throws InvalidDataFormatException {
        assert(_weights != null);
        assert(_network != null);
        assert(_weights.getNodeName().equals(WEIGHTS_NODE_NAME));
        assert(_weights.getNodeType() == Node.ELEMENT_NODE);
        
        Element weights_elm = (Element)_weights;
        NodeList rows = weights_elm.getElementsByTagName(ROW_WEIGHTS_NODE_NAME);
        if(rows.getLength() != _network.getRowCount() - 1)
            throw new InvalidDataFormatException("Number of rows in the file does not match number of rows in the network");
        
        // zpracuj vahy jednotlive vrstvy. Vahy jsou ukladany pro vsechny
        // neurony mimo ty ve vstupni vrstve (nejsou nikam napojene)
        for(int row = 0; row < rows.getLength(); row++) {
            // pri vyhledavani site vynechj vstupni vrstvu
            NeuronRow current_row = _network.getRow(row + 1);
            
            Element row_elm = (Element)rows.item(row);
            NodeList neurons = row_elm.getElementsByTagName(NEURON_WEIGHTS_NODE_NAME);
            if(neurons.getLength() != current_row.getSize())
                throw new InvalidDataFormatException("Number of neurons in a row (" + (row + 1) + ") does not match number of neurons in the file for that row");
            
            // zpracuj vahy pro jednotlive neurony ve vrstve
            for(int neuron = 0; neuron < neurons.getLength(); neuron++) {
                Neuron currentNeuron = current_row.getNeuron(neuron);
                
                Element neuron_elm = (Element)neurons.item(neuron);
                
                // prah je ulozeny zvlast jako atribut... zkontroluj, jestli je pritomen
                if(!neuron_elm.hasAttribute(THRESHOLD_ATTRIBUTE_NAME))
                    throw new InvalidDataFormatException("A neuron does not have it's threshold specified");
                
                try {
                    double threshold = Double.parseDouble(neuron_elm.getAttribute(THRESHOLD_ATTRIBUTE_NAME));
                    currentNeuron.setThreshold(threshold);
                    
                    double[] weights = currentNeuron.getWeights();
                    StreamTokenizer tokenizer = new StreamTokenizer(new StringReader(neuron_elm.getTextContent()));
                    tokenizer.parseNumbers();
                    
                    for(int i = 0; i < weights.length; i++) {
                        tokenizer.nextToken();
                        // kontrola, ze jsme nenarazili na konec vstupu nebo na
                        // nejake neocekavane objekty moc brzy
                        if(StreamTokenizer.TT_NUMBER != tokenizer.ttype)
                            throw new InvalidDataFormatException("Number expected when parsing weight data");
                        weights[i] = tokenizer.nval;
                    }
                    // kontrola, ze toho ve vstupnim proudu neni vic, nez by melo
                    if(StreamTokenizer.TT_EOF != tokenizer.nextToken())
                        throw new InvalidDataFormatException("There are more numbers in the file than weights for current neuron");
                }
                catch(NumberFormatException nf_err) {
                    // doslo k nejake chybe pri prevodu textove reprezentace na cisla
                    throw new InvalidDataFormatException("Neuron threshold is not a valid number");
                }
                catch(IOException io_err) {
                    // nejaka IO chyba pri cteni vah... cte se z retezce, takze
                    // by toto nemelo nikdy nastat
                    throw new InvalidDataFormatException("There was an error parsing weights data");
                }
            }
        }
    }
}
