/*
 * lNeuron.java
 *
 * Created on 4. kvten 2005, 15:18
 */

package cz.matfyz.sykora.nngui.network;

/**
 * Tda implementujc neuron jako vpoetn jednotku st. Do <i>Neuron</i>
 * by mla bt soustedna pouze funknost tkajc se samotnho vpotu st,
 * adaptan algoritmy jsou implementovny pomoc jinch td tak, aby bylo
 * mon je snadno mnit a nahrazovat
 *  @author Ondej Skora [ondrasej@centrum.cz]
 */
public class Neuron {
    /**
     * Excitace neuronu (souet vstup pensobench vahami), pot se pi
     * vpotu hodnot st, z nj se pak pot vstupn hodnota neuronu
     *  @see Neuron#calculate()
     */
    private double excitation       = 0.0;
    /**
     * Vstupn hodnota neuronu (excitace prohnan prahovou funkc), pot se
     * z exicitace pi vpotu hodnot st
     *  @see Neuron#excitation
     *  @see Neuron#calculate
     */
    private double output           = 0.0;
    /**
     * Rozdl mezi oekvanm vstupem a vstupem skutenm, tato hodnota se
     * v sti aproximuje pi uen zptnm enm chyby
     *  @see Neuron#output
     */
    private double difference       = 0.0;
    /**
     * Odkaz na objekt zastreujc neuronovou s, pro kterou tento neuron
     * pot.
     */
    private Network network;
    /**
     * Pole s odkazy na (vechny) neurony v pedchoz vrstv. Toto pole sdlej
     * vechny neurony v jedn vrstv.
     *  @see Neuron#weights
     */
    private Neuron[] predecessors;
    /**
     * Pole vah mezi tmto neuronem a neurony z pedchoz vrstvy. Vha pro dan
     * neuron z <i>predecessors</i> m stejn index jako neuron, pro kter je
     * vha urena
     *  @see Neuron#predecessors
     *  @see Neuron#threshold
     *  @see Neuron#setWeights
     */
    private double[] weights;
    /**
     * Rozdly vah akumulovan bhem algoritmu zptnho en. Rozdl pro danou
     * vhu m v tomto poli stejn index jako vha v poli <i>weights</i>
     *  @see #processDifference
     *  @see AdaptationStrategy
     *  @see AdaptationStrategy#adaptNeuron
     */
    private double[] weightDelta;
    /**
     * Prh pro tento neuron
     *  @see #getThreshold
     *  @see #setThreshold
     *  @see #thresholdDelta
     */
    private double threshold        = 0.0;
    /**
     * Rozdl prahu tohoto neuronu akumulovan bhem algoritmu zptnho en.
     *  @see #threshold
     *  @see #addThresholdDelta
     *  @see #getThresholdDelta
     *  @see #setThresholdDelta
     */
    private double thresholdDelta   = 0.0;
    
    /**
     * Akumulace rozdl oekvanch a skutench hodnot tohoto neuronu. 
     *  @param _difference hodnota, kter m bt pitena k aktulnmu rozdlu
     *              oekvanch a skutench hodnot.
     *  @see #difference
     *  @see #processDifference
     */
    protected void addDifference(double _difference) {
        difference += _difference;
    }
    
    /**
     * Akumulace rozdl prahu tohoto neuronu
     *  @param _difference hodnota, kter m bt pitena k aktulnmu rozdlu
     *              prahu tohoto neuronu
     *  @see #thresholdDelta
     *  @see #processDifference
     */
    protected void addThresholdDelta(double _difference) {
        thresholdDelta += _difference;
    }
    
    /**
     * Dopedn vpoet v tomto uzlu neuronov st, pedpokld, e v pedchoz
     * vrstv ji vpoet probhl a bere si z n vstupn hodnoty
     *  @return vypotan vstupn hodnotu neuronu
     */
    public double calculate() {
        assert(weights != null);
        assert(predecessors != null);
        assert(predecessors.length == weights.length);

        // vypocti aktivacni hodnotu tohoto neuronu...
        excitation = getThreshold();
        for (int i = 0; i < predecessors.length; i++) {
            excitation += weights[i] * predecessors[i].getOutput();
        }
        
        // ...a transformuj ji na vystupni hodnotu
        setOutput(network.calculateTransition(excitation));
        return getOutput();
    }
    
    /**
     * Zjitn odchylky skuten hodnoty neuronu od oekvan hodnoty
     *  @return odchylka mezi skutenou a oekvanou hodnotou neuronu
     *  @see #difference
     *  @see #processDifference
     */
    public final double getDifference() {
        return difference;
    }
    
    /**
     * Zjitn vstupn hodnoty neuronu
     *  @return vstupn hodnota neuronu
     *  @see #output
     *  @see #setOutput
     *  @see #calculate
     */
    public final double getOutput() {
        return output;
    }
    
    /**
     * Zjitn potu neuron v pedchoz vrstv (resp. potu pedchdc tohoto
     * neuronu.
     *  @return poet pedchdc tohoto neuronu
     *  @see #predecessors
     */
    public final int getPredecessorssCount() {
        assert(predecessors != null);
        
        return predecessors.length;
    }
    
    /**
     * Zjitn aktulnho prahu tohoto neuronu.
     *  @return prh tohoto neuronu
     *  @see #threshold
     *  @see #setThreshold
     */
    public final double getThreshold() {
        return threshold;
    }
    
    /**
     * Zjitn aktulnho rozdlu prahu tohoto neuronu
     *  @return aktuln rozdl prahu tohoto neuronu
     *  @see #thresholdDelta
     */
    public final double getThresholdDelta() {
        return thresholdDelta;
    }
    
    /**
     * Pstup k poli s vahami mezi tmto neuronem a neurony z pedchoz vrstvy
     *  @return pole obsahujc vhy mezi tmto neuronem a neurony pedchoz
     *          vrstvy. Pozor, tato funkce nevytv kopii pole a proto vechny
     *          zmny proveden ve vrcenm poli se projev pi dalch
     *          vpotech tohoto neuronu
     *  @see #getWeightDelta
     *  @see #weights
     *  @see #weightDelta
     */
    public final double[] getWeights() {
        return weights;
    }
    
    /**
     * Pstup k poli s rozdly vah mezi tmro neuronem a neurony pedchoz
     * vrstvy.
     *  @return pole obsahujc vhy mezi tmto neuronem a neurony pedchoz
     *          vrstvy. Pozor, tato funkce nevytv kopii pole a proto se
     *          vechny zmny proveden ve vrcenm poli mohou projevit pi
     *          dalch vpotech
     *  @see #getWeights
     *  @see #weightDelta
     *  @see #weights
     */
    public final double[] getWeightDelta() {
        return weightDelta;
    }
    
    /**
     * Inicializace vah spoj jdoucch z pedchoz vrstvy do tohoto neuronu
     * a prahov hodnoty malmi nhodnmi sly.
     *  @see #weights
     *  @see #threshold
     */
    protected void initializeRandomWeights() {
        assert(weights != null);
        
        for (int i = 0; i < weights.length; i++)
            weights[i] = network.getSmallRandomNumber();
        threshold = network.getSmallRandomNumber();
    }
    
    /**
     * Zpracovn chyby na tomto neuronu a jej distribuce do neuron pedchoz
     * vrstvy st. Implementuje zkladn proces uen zptnovazebn st
     *  @see #difference
     *  @see #thresholdDelta
     *  @see #weightDelta
     *  @see #weights
     */
    public void processDifference() {
        assert(weights != null);
        assert(predecessors != null);
        assert(weights.length == predecessors.length);
        
        // prenasob derivaci prechodove funkce pro tento neuron (nasobi se jenom
        // jednou na konci vypoctu, stejne to je komutativni operace)
        difference = difference * network.calculateTransitionDerivation(excitation);
        
        for (int i = 0; i < weights.length; i++) {
            assert(predecessors[i] != null);
            // podle vah distribuuji rozdil mezi ocekavanym a skutencym vystupem
            // do neuronu predchozi vrstvy
            predecessors[i].addDifference(getDifference() * weights[i]);
            
            // nastav zmenu vah - prenasobenou vystupem prislusneho neronu z
            // predchozi vrstvy (zbytek vypoctu derivace chybove funkce)
            weightDelta[i] += getDifference() * predecessors[i].getOutput();
            // uprav rozdil prahu tohoto neuronu
            addThresholdDelta(getDifference());
        }
    }
    
    /**
     * Nastavan rozdlu mezi skutenm a oekvanm vstupem
     *  @param _difference nastavovan hodnota rozdlu
     */
    protected void setDifference(double _difference) {
        difference = _difference;
    }
    
    /**
     * Nastaven vstupn hodnoty tohoto neuronu
     *  @param _output piazovan vstupn hodnota
     */
    protected void setOutput(double _output) {
        output = _output;
    }
    
    /**
     * Piazen vektoru pedchdc neuronu. Tento vektor se v prbhu ivota
     * neuronu nemn, me bt sdlen dalmi neurony v sti. Je to prv
     * seznam neuron z pedchoz vrstvy.
     *  @param _preds seznam pedchdc tohoto neuronu
     */
    protected void setPredecessors(Neuron[] _preds) {
        predecessors = _preds;
    }
    /**
     * Nastaven prahu pro tento neuron.
     *  @param _thresh nov hodnota prahu pro tento neuron
     *  @see #threshold
     */
    public final void setThreshold(double _thresh) {
        threshold = _thresh;
    }
    
    /**
     * Nastaven rozdlu prahu pro tento neuron
     *  @param _delta rozdl prahu pro tento neuron
     *  @see #thresholdDelta
     */
    public final void setThresholdDelta(double _delta) {
        thresholdDelta = _delta;
    }
    
    /**
     * Piazen vhovho vektoru jednomu neuronu. Tento vektor se nekopruje,
     * navc se v nm pozdji provdj pravy bhem adaptace st, proto je
     * dleit, aby se nepouval nkde jinde k njakm dalm elm
     *  @param _weights vhov vektor piazen tomuto neuronu
     */
    protected void setWeights(double[] _weights) {        
        weights = _weights;
        if(null != weights)
            weightDelta = new double[_weights.length];
        else
            weightDelta = null;
    }
    
    /**
     * Konstruktor, inicializuje neuron pro pouit v zadan sti. Je skryt
     * tak, aby uivatel mohl strukturu st mnit pouze "bezpenmi" funkcemi
     * pro editaci struktury
     *  @param _net s, pro kterou neuron inicializujeme
     */
    protected Neuron(Network _net) {
        network = _net;
    }
}
